import logging
import os
import shutil
import tempfile
import traceback
from dataclasses import dataclass
from enum import Enum, auto

from stringcase import lowercase, titlecase
from video_utils import Video

from .colour import colour
from .ffmpeg_converter import FFmpegConverter
from .settings import AudioSettings, VideoSettings

log = logging.getLogger()


class Status(Enum):
    # Successful conversion
    CONVERTED = auto()

    # The file would be converted, if we weren't running in dry-run mode
    WOULD_CONVERT = auto()

    # If a file is already converted with renaming
    # (e.g. appending the output codec) and this is that converted file
    ALREADY_PROCESSED = auto()

    # The file is already using the target format, can be overridden with --force
    IN_DESIRED_FORMAT = auto()

    # The file is below the minimum size and will not be processed
    BELOW_MINIMUM_SIZE = auto()

    FAILED = auto()

    def __str__(self):
        status_text = titlecase(lowercase(self.name))
        if self not in (Status.FAILED, Status.WOULD_CONVERT, Status.CONVERTED):
            return f"SKIPPING: {status_text}"
        return status_text

    def colour(self):
        c = "green"
        if self not in (Status.WOULD_CONVERT, Status.CONVERTED):
            c = "blue"
        if self == Status.FAILED:
            c = "red"
        return colour(c, str(self))


@dataclass
class VideoProcessor:
    video: Video
    video_settings: VideoSettings
    audio_settings: AudioSettings
    container: str
    force: bool = False
    dry_run: bool = False
    in_place: bool = False
    extra_ffmpeg_input_args: str = ""
    extra_ffmpeg_output_args: str = ""
    temp_directory: str = None  # type: ignore
    minimum_size_b: int = 0  # Minimum file size in megabytes to process

    def _create_temp_file(self):
        return tempfile.NamedTemporaryFile(
            dir=self.temp_directory, suffix=f".{self.container}"
        )

    def __str__(self):
        codec_name = (
            self.video.codec.pretty_name if self.video.codec is not None else "Unknown"
        )
        return f"Video: {self.video.full_path}, format: {codec_name}, quality: {self.video.quality}"

    def process(self):
        if self._is_below_minimum_size():
            return Status.BELOW_MINIMUM_SIZE

        if self.video.codec == self.video_settings.codec:
            log.debug(f"'{self.video.name}' is already in the desired format")
            if not self.force:
                return Status.IN_DESIRED_FORMAT
            log.debug("Forcing conversion anyway (--force is enabled)")

        if self.already_processed():
            return Status.ALREADY_PROCESSED

        with self._create_temp_file() as self.temp_file:
            try:
                converter = FFmpegConverter(
                    source_file_path=self.video.full_path,
                    destination_file_path=self.temp_file.name,
                    extra_ffmpeg_input_args=self.extra_ffmpeg_input_args,
                    extra_ffmpeg_output_args=self.extra_ffmpeg_output_args,
                    video_settings=self.video_settings,
                    audio_settings=self.audio_settings,
                    dry_run=self.dry_run,
                )
                converter.process()
                self._move_output_video()
                if self.dry_run:
                    return Status.WOULD_CONVERT
                return Status.CONVERTED
            except Exception as e:
                log.error(
                    colour(
                        "red", f"Failed to convert {self.video.full_path}. Exception:"
                    )
                )
                log.error(e)
                traceback.print_exc()
                return Status.FAILED

    def _is_below_minimum_size(self):
        if self.minimum_size_b and self.video.size_b:
            if self.video.size_b < self.minimum_size_b:
                return True
        return False

    def already_processed(self):
        renamed_path = self.renamed_path()
        if os.path.exists(renamed_path):
            log.debug(
                f"File '{self.video.name}' appears to have already been converted to {renamed_path} exists. Skipping"
            )
            return True

        split_filename = os.path.splitext(self.video.name)
        codec_name = (
            self.video_settings.codec.pretty_name
            or self.video_settings.get_ffmpeg_codec()
        )
        if split_filename[0].endswith(codec_name):
            log.debug(
                f"File '{self.video.name}' already matches the renaming format. Skipping"
            )
            return True

        return False

    def renamed_path(self):
        split_filename = os.path.splitext(self.video.full_path)
        codec_name = self.video_settings.codec.pretty_name
        return f"{split_filename[0]} - {codec_name}.{self.container}"

    def in_place_file_path(self):
        split_filename = os.path.splitext(self.video.full_path)
        return f"{split_filename[0]}.{self.container}"

    def _move_output_video(self):
        log.debug("Moving file from temporary storage back to original folder")
        if not self.dry_run:
            shutil.move(self.temp_file.name, self.renamed_path())
        if self.in_place:
            if self.dry_run:
                log.info(
                    colour(
                        "blue",
                        f"DRY-RUN: Would replace original file {self.video.full_path}",
                    )
                )
                return
            print(f"Replacing original file {self.video.full_path}")
            os.remove(self.video.full_path)
            shutil.move(self.renamed_path(), self.in_place_file_path())



import logging

import click
from prettytable import PrettyTable
from video_utils import Codec

from convert_videos.util import check_hardware_acceleration_support

from .processor import AudioSettings, Processor, VideoSettings

CONTEXT_SETTINGS = dict(help_option_names=["--help", "-h"])
LOG_FORMATTER = logging.Formatter("%(asctime)s - %(message)s", "%Y-%m-%d %H:%M:%S")


def configure_logger(verbose):
    global log
    log_level = logging.INFO
    if verbose:
        log_level = logging.DEBUG

    logging.basicConfig(level=log_level)
    log = logging.getLogger()
    log.handlers[0].setFormatter(LOG_FORMATTER)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("directories", nargs=-1, type=click.Path(exists=True), required=True)
@click.option(
    "--in-place",
    "-i",
    is_flag=True,
    help="Replace the original files instead of appending the new codec name",
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force conversion even if the format of the file already matches the desired format",
)
@click.option(
    "--video-codec",
    default="HEVC",
    show_default=True,
    help="A target video codec. Supported codecs: HEVC, AVC",
)
@click.option(
    "--quality",
    "-q",
    type=int,
    default=24,
    show_default=True,
    help="The quantizer quality level to use.",
)
@click.option(
    "--preset", "-p", default="medium", show_default=True, help="FFmpeg preset to use."
)
@click.option(
    "--width",
    "-w",
    type=int,
    help="Specify a new width to enable resizing of the video",
)
@click.option(
    "--extra-input-args",
    default="",
    help="Specify any extra arguments you would like to pass to FFMpeg input here",
)
@click.option(
    "--extra-output-args",
    default="",
    help="Specify any extra arguments you would like to pass to FFMpeg output here",
)
@click.option(
    "--audio-codec", default="AAC", show_default=True, help="A target audio codec"
)
@click.option(
    "--audio-channels",
    default=2,
    show_default=True,
    type=int,
    help="The number of channels to mux sound in to",
)
@click.option(
    "--audio-bitrate",
    default=160,
    show_default=True,
    type=int,
    help="The bitrate to use for the audio codec",
)
@click.option(
    "--temp-dir",
    help="Specify a temporary directory to use during conversions instead of the system default",
)
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose log output")
@click.option(
    "--container",
    default="mkv",
    show_default=True,
    help="Specify a video container to convert the videos in to",
)
@click.option("--dry-run", is_flag=True, help="Do not make actual changes")
@click.option(
    "--encoder",
    type=click.Choice(["auto-detect", "software", "nvidia", "intel"]),
    default="auto-detect",
    show_default=True,
    help="Optionally use a hardware encoder to speed things up.",
)
@click.option("--audio-language", help="Only include audio streams in this language")
@click.option(
    "--subtitle-language", help="Only include subtitle streams in this language"
)
@click.option(
    "--minimum-size",
    type=int,
    default=0,
    help="Skip processing videos if they are below the specified size",
)
def main(
    directories,
    force,
    in_place,
    video_codec,
    quality,
    preset,
    width,
    extra_input_args,
    extra_output_args,
    audio_codec,
    audio_channels,
    audio_bitrate,
    temp_dir,
    verbose,
    container,
    dry_run,
    encoder,
    audio_language,
    subtitle_language,
    minimum_size,
):
    configure_logger(verbose)

    if encoder == "auto-detect":
        log.info("Auto detecting hardware acceleration support")
        hardware_support = check_hardware_acceleration_support()
        encoder = "software"
        if hardware_support["intel_quicksync"]:
            encoder = "intel"
        if hardware_support["nvidia_nvenc"]:
            encoder = "nvidia"
        log.info(f"Using encoder: {encoder}")

    video_settings = VideoSettings(
        codec=Codec(video_codec),
        quality=quality,
        preset=preset,
        width=width,
        encoder=encoder,
        subtitle_language=subtitle_language,
    )
    audio_settings = AudioSettings(
        codec=Codec(audio_codec),
        channels=audio_channels,
        bitrate=audio_bitrate,
        language=audio_language,
    )

    results = []
    for directory in directories:
        results += Processor(
            directory=directory,
            force=force,
            video_settings=video_settings,
            audio_settings=audio_settings,
            in_place=in_place,
            extra_ffmpeg_input_args=extra_input_args,
            extra_ffmpeg_output_args=extra_output_args,
            temp_directory=temp_dir,
            container=container,
            dry_run=dry_run,
            minimum_size_b=minimum_size * 1024 * 1024,
        ).start()

    _print_conversion_results(results)


def _print_conversion_results(results):
    table = PrettyTable(["Video", "Original Codec", "Current Size", "Status"])

    for result in results:
        codec = result["video"].codec.pretty_name
        if codec is None:
            codec = "Unknown"
        table.add_row(
            [
                result["video"].name,
                codec,
                str(result["video"].size_b // (1024 * 1024)) + " MB",
                result["status"].colour(),
            ]
        )
    print(table)



from colorama import Fore, Style


def colour(colour, message):
    colours = {
        "green": Fore.GREEN,
        "blue": Fore.BLUE,
        "red": Fore.RED,
        "yellow": Fore.YELLOW,
    }
    return colours[colour] + str(message) + Style.RESET_ALL



import pytest
from mock import mock_open, patch
from video_utils import Codec

from convert_videos import ffmpeg_converter
from convert_videos.ffmpeg_converter import FFmpegConverter
from convert_videos.settings import AudioSettings, VideoSettings


@pytest.fixture
@patch("os.path.exists", return_value=True)
@patch("os.path.isfile", return_value=True)
@patch("os.access", return_value=True)
@patch("os.path.isdir")
@patch("os.remove")
def target(mock_remove, mock_isdir, mock_access, mock_isfile, mock_exists):
    audio_settings = AudioSettings(codec=Codec("AAC"), channels=2, bitrate=120)

    video_settings = VideoSettings(
        codec=Codec("HEVC"), quality=25, preset="slow", encoder="software"
    )

    return FFmpegConverter(
        source_file_path="/asdf/foo/bar.mkv",
        video_settings=video_settings,
        audio_settings=audio_settings,
        destination_file_path="/asdf/temp/path.mkv",
        extra_ffmpeg_input_args="",
        extra_ffmpeg_output_args="",
        dry_run=False,
    )


@patch.object(FFmpegConverter, "_generate_ffmpeg_settings", return_value="12345")
@patch.object(ffmpeg_converter, "ffmpy")
def test_process(mock_ffmpy, mock_settings, target):
    target.process()
    mock_ffmpy.FFmpeg.assert_called_with(
        inputs={"/asdf/foo/bar.mkv": "12345"}, outputs={"/asdf/temp/path.mkv": "12345"}
    )
    mock_ffmpy.FFmpeg().run.assert_called()


@patch.object(FFmpegConverter, "_generate_ffmpeg_settings", return_value="12345")
@patch.object(ffmpeg_converter, "ffmpy")
def test_process_dryrun(mock_ffmpy, mock_settings, target):
    target.dry_run = True
    target.process()
    mock_ffmpy.FFmpeg.assert_called_with(
        inputs={"/asdf/foo/bar.mkv": "12345"}, outputs={"/asdf/temp/path.mkv": "12345"}
    )
    mock_ffmpy.FFmpeg().run.assert_not_called()


def test_generate_ffmpeg_output_settings(target):
    result = target._generate_ffmpeg_settings("output")
    assert (
        result
        == "-y -threads 0 -map 0:v -map 0:s? -c:s copy -vcodec libx265 -preset slow -crf 25 -strict -2 -map 0:a -acodec aac -ab 120k -ac 2 "
    )


def test_extra_output_args(target):
    target.extra_ffmpeg_output_args = "-foo"
    result = target._generate_ffmpeg_settings("output")
    assert (
        result
        == "-y -threads 0 -map 0:v -map 0:s? -c:s copy -vcodec libx265 -preset slow -crf 25"
        + " -strict -2 -map 0:a -acodec aac -ab 120k -ac 2 -foo"
    )


def test_generate_ffmpeg_input_settings(target):
    result = target._generate_ffmpeg_settings("input")
    assert result == " "


def test_generate_ffmpeg_input_settings_nvidia_hw(target):
    target.video_settings.encoder = "nvidia"
    result = target._generate_ffmpeg_settings("input")
    assert result == " -hwaccel cuda"


def test_generate_ffmpeg_input_settings_intel_hw(target):
    target.video_settings.encoder = "intel"
    result = target._generate_ffmpeg_settings("input")
    assert result == " -hwaccel qsv -hwaccel_output_format qsv"


def test_extra_input_args(target):
    target.extra_ffmpeg_input_args = "-foo"
    result = target._generate_ffmpeg_settings("input")
    assert result == " -foo"


@patch("os.path.exists", return_value=True)
@patch("os.path.isfile", return_value=True)
@patch("os.access", return_value=True)
@patch("os.path.isdir")
@patch("os.remove")
def test_validate_destination_happy_path_exists(
    mock_remove, mock_isdir, mock_access, mock_isfile, mock_exists, target
):
    result = target._validate_destination()
    assert result is True


@patch("os.path.exists", return_value=True)
@patch("os.path.isfile", return_value=True)
@patch("os.access", return_value=False)
@patch("os.path.isdir")
@patch("os.remove")
def test_validate_destination_unwritable(
    mock_remove, mock_isdir, mock_access, mock_isfile, mock_exists, target
):
    with pytest.raises(IOError):
        target._validate_destination()


@patch("os.path.exists", return_value=True)
@patch("os.path.isfile", return_value=False)
@patch("os.access")
@patch("os.path.isdir", return_falue=True)
@patch("os.remove")
def test_validate_destination_is_dir(
    mock_remove, mock_isdir, mock_access, mock_isfile, mock_exists, target
):
    with pytest.raises(IOError):
        target._validate_destination()


@patch("os.path.exists", return_value=False)
@patch("os.path.isfile")
@patch("os.access")
@patch("os.path.isdir")
@patch("os.remove")
def test_validate_destination_doesnt_exist(
    mock_remove, mock_isdir, mock_access, mock_isfile, mock_exists, target
):
    m = mock_open()
    with patch("builtins.open", m):
        result = target._validate_destination()
        assert result is True


@patch("os.path.exists", return_value=False)
@patch("os.path.isfile")
@patch("os.access")
@patch("os.path.isdir")
@patch("os.remove")
def test_validate_destination_invalid_path(
    mock_remove, mock_isdir, mock_access, mock_isfile, mock_exists, target
):
    m = mock_open()
    m.side_effect = FileNotFoundError
    with patch("builtins.open", m):
        with pytest.raises(FileNotFoundError):
            target._validate_destination()



from convert_videos.settings import AudioSettings, VideoSettings
from video_utils import Codec


def test_audio_settings():
    target = AudioSettings(Codec("AAC"), 2, 128)
    assert str(target) == " -map 0:a -acodec aac -ab 128k -ac 2"


def test_audio_copy():
    target = AudioSettings(Codec("copy"), 2, 128)
    assert str(target) == " -map 0:a -acodec copy"


def test_audio_language_filter_fullname():
    target = AudioSettings(Codec("AAC"), 2, 128, "english")
    assert str(target) == " -map 0:a:m:language:eng -acodec aac -ab 128k -ac 2"


def test_audio_language_filter_shortcode():
    target = AudioSettings(Codec("AAC"), 2, 128, "jpn")
    assert str(target) == " -map 0:a:m:language:jpn -acodec aac -ab 128k -ac 2"


def test_video_settings_avc():
    target = VideoSettings(Codec("AVC"), 25, "slow")
    assert (
        str(target) == " -map 0:v -map 0:s? -c:s copy -vcodec h264 -preset slow -crf 25"
    )


def test_video_settings_avc_nvidia():
    target = VideoSettings(Codec("AVC"), 25, "slow", encoder="nvidia")
    assert (
        str(target)
        == " -map 0:v -map 0:s? -c:s copy -vcodec h264_nvenc -preset slow -rc constqp -qp 25 -b:v 0"
    )


def test_video_settings_avc_intel():
    target = VideoSettings(Codec("AVC"), 25, "slow", encoder="intel")
    assert (
        str(target)
        == " -map 0:v -map 0:s? -c:s copy -vcodec h264_qsv -preset slow -global_quality 25 -look_ahead 1 -g:v 120"
    )


def test_video_settings_hevc():
    target = VideoSettings(Codec("HEVC"), 25, "slow", None)  # type: ignore
    assert (
        str(target)
        == " -map 0:v -map 0:s? -c:s copy -vcodec libx265 -preset slow -crf 25 -strict -2"
    )


def test_video_settings_hevc_nvidia():
    target = VideoSettings(Codec("HEVC"), 25, "slow", encoder="nvidia")
    assert (
        str(target)
        == " -map 0:v -map 0:s? -c:s copy -vcodec hevc_nvenc -preset slow -rc constqp -qp 25 -b:v 0 -strict -2"
    )


def test_video_settings_hevc_intel():
    target = VideoSettings(Codec("HEVC"), 25, "slow", encoder="intel")
    assert (
        str(target)
        == " -map 0:v -map 0:s? -c:s copy -vcodec hevc_qsv -preset slow -global_quality 25 -g:v 120 -strict -2"
    )


def test_video_settings_width():
    target = VideoSettings(
        codec=Codec("AVC"), quality=25, preset="slow", width=720, encoder="software"
    )
    assert (
        str(target)
        == " -map 0:v -map 0:s? -c:s copy -vcodec h264 -preset slow -crf 25 -vf scale=720:-2"
    )


def test_video_copy():
    target = VideoSettings(Codec("copy"), 25, "slow")
    assert str(target) == " -map 0:v -map 0:s? -c:s copy -vcodec copy"


def test_video_language_filter_fullname():
    target = VideoSettings(Codec("AVC"), 25, "slow", subtitle_language="english")
    assert (
        str(target)
        == " -map 0:v -map 0:s:m:language:eng? -c:s copy -vcodec h264 -preset slow -crf 25"
    )


def test_video_language_filter_shortcode():
    target = VideoSettings(Codec("AVC"), 25, "slow", subtitle_language="jpn")
    assert (
        str(target)
        == " -map 0:v -map 0:s:m:language:jpn? -c:s copy -vcodec h264 -preset slow -crf 25"
    )



from convert_videos import processor, Processor, AudioSettings, VideoSettings
from convert_videos.video_processor import Status
from video_utils import Codec, Video
import pytest
from mock import patch, NonCallableMagicMock
from os import path
import pickle


@pytest.fixture
def target():
    audio_settings = AudioSettings(codec=Codec("AAC"), channels=2, bitrate=120)

    video_settings = VideoSettings(
        codec=Codec("HEVC"),
        quality=25,
        preset="slow",
        encoder="software",
    )

    return Processor(
        directory="/tmp/foo",
        force=False,
        video_settings=video_settings,
        audio_settings=audio_settings,
        in_place=False,
        container="mkv",
    )


@pytest.fixture
def file_map_fixture():
    current_dir = path.dirname(path.abspath(__file__))
    test_data_dir = path.join(current_dir, "testData")
    filemap_file = path.join(test_data_dir, "filemap.pickle")
    with open(filemap_file, "rb") as f:
        return pickle.load(f)


@patch.object(Processor, "_load_file_map", autospec=True)
@patch.object(Processor, "_convert_all", autospec=True)
def test_processor_start(mock_convert_all, mock_load_file_map, target):
    target.results = "foo"
    result = target.start()
    mock_convert_all.assert_called()
    mock_load_file_map.assert_called()
    assert result == "foo"


@patch.object(processor, "FileMap")
def test_load_file_map(mock_file_map, target):
    target._load_file_map()
    mock_file_map.assert_called_with("/tmp/foo")
    mock_file_map().load.assert_called()


@patch.object(processor, "VideoProcessor", autospec=True)
def test_get_video_processor(mock_video_processor, target):
    result = target._get_video_processor(Video("bar.mkv", "/tmp/foo"))
    mock_video_processor.assert_called()
    assert isinstance(result, NonCallableMagicMock)


@patch.object(Processor, "_convert_files_in_directory", return_value=["some-response"])
def test_convert_all(mock_convert_files_in_directory, target, file_map_fixture):
    target._file_map = file_map_fixture
    target._convert_all()
    assert mock_convert_files_in_directory.call_count == 2
    mock_convert_files_in_directory.assert_any_call(
        "/Users/jdray/git/home/convert_videos/tests/testData/foo"
    )
    mock_convert_files_in_directory.assert_any_call(
        "/Users/jdray/git/home/convert_videos/tests/testData/bar"
    )

    # Confirm we return the results from convert_files_in_directory in a list
    assert target.results == ["some-response", "some-response"]


@patch.object(Processor, "_get_video_processor")
def test_convert_files_in_directory_status_passthrough(
    mock_get_video_processor, target, file_map_fixture
):
    mock_get_video_processor().process.return_value = Status.FAILED
    target._file_map = file_map_fixture
    response = target._convert_files_in_directory(
        "/Users/jdray/git/home/convert_videos/tests/testData/foo"
    )

    failures = [x for x in response if x["status"] == Status.FAILED]

    assert len(failures) == 6


@patch.object(Processor, "_get_video_processor")
def test_convert_files_in_directory_below_minimum_size(
    mock_get_video_processor, target, file_map_fixture
):
    mock_get_video_processor().process.return_value = Status.BELOW_MINIMUM_SIZE
    target.minimum_size = 100
    target._file_map = file_map_fixture
    response = target._convert_files_in_directory(
        "/Users/jdray/git/home/convert_videos/tests/testData/foo"
    )

    below_minimum_size = [
        x for x in response if x["status"] == Status.BELOW_MINIMUM_SIZE
    ]

    assert len(below_minimum_size) == 6



import pytest
from mock import Mock, patch
from video_utils import Codec, Video

from convert_videos import video_processor
from convert_videos.settings import AudioSettings, VideoSettings
from convert_videos.video_processor import Status, VideoProcessor


def mock_temp_file(name):
    mock_file = Mock()
    mock_file.name = name
    mock_file.__enter__ = Mock(return_value=mock_file)
    mock_file.__exit__ = Mock(return_value=None)
    return mock_file


@pytest.fixture
def target():
    audio_settings = AudioSettings(codec=Codec("AAC"), channels=2, bitrate=120)

    video_settings = VideoSettings(
        codec=Codec("HEVC"), quality=25, preset="slow", encoder="software"
    )

    return VideoProcessor(
        video=Video("bar.mkv", "/asdf/foo"),
        video_settings=video_settings,
        audio_settings=audio_settings,
        container="asdf",
        in_place=False,
        dry_run=False,
    )


@patch("tempfile.NamedTemporaryFile", autospec=True)
def test_create_temp_file_default(mock_NamedTemporaryFile, target):
    target._create_temp_file()
    mock_NamedTemporaryFile.assert_called_with(dir=None, suffix=".asdf")


@patch("tempfile.NamedTemporaryFile", autospec=True)
def test_create_temp_file_new_dir(mock_NamedTemporaryFile, target):
    target.temp_directory = "/foo/bar"
    target._create_temp_file()
    mock_NamedTemporaryFile.assert_called_with(dir="/foo/bar", suffix=".asdf")


@patch.object(video_processor, "FFmpegConverter")
@patch.object(VideoProcessor, "_move_output_video")
@patch.object(VideoProcessor, "_create_temp_file")
def test_process(
    mock_create_temp_file, mock_move_output_video, mock_ffmpeg_converter, target
):
    mock_create_temp_file.return_value = mock_temp_file("/foo")
    target.process()
    mock_ffmpeg_converter.assert_called_with(
        audio_settings=target.audio_settings,
        video_settings=target.video_settings,
        extra_ffmpeg_input_args="",
        extra_ffmpeg_output_args="",
        destination_file_path="/foo",
        source_file_path="/asdf/foo/bar.mkv",
        dry_run=False,
    )
    mock_ffmpeg_converter().process.assert_called()
    mock_move_output_video.assert_called()


@patch.object(video_processor, "FFmpegConverter")
@patch.object(VideoProcessor, "_move_output_video")
@patch.object(VideoProcessor, "_create_temp_file")
def test_process_dry_run(
    mock_create_temp_file, mock_move_output_video, mock_ffmpeg_converter, target
):
    mock_create_temp_file.return_value = mock_temp_file("/foo")
    target.dry_run = True
    target.process()
    mock_ffmpeg_converter.assert_called_with(
        audio_settings=target.audio_settings,
        video_settings=target.video_settings,
        extra_ffmpeg_input_args="",
        extra_ffmpeg_output_args="",
        destination_file_path="/foo",
        source_file_path="/asdf/foo/bar.mkv",
        dry_run=True,
    )
    mock_ffmpeg_converter().process.assert_called()
    mock_move_output_video.assert_called()


@patch.object(VideoProcessor, "renamed_path")
@patch("os.path.exists", return_value=True)
def test_already_processed_renamed_file_exists(mock_exists, mock_renamed_path, target):
    result = target.already_processed()
    assert result is True


@patch.object(VideoProcessor, "renamed_path")
@patch("os.path.exists", return_value=False)
def test_already_processed_container_matches(mock_exists, mock_renamed_path, target):
    target.video.name = "bar - HEVC.mkv"
    result = target.already_processed()
    assert result is True


@patch.object(VideoProcessor, "renamed_path")
@patch("os.path.exists", return_value=False)
def test_already_processed_false(mock_exists, mock_renamed_path, target):
    result = target.already_processed()
    assert result is False


def test_renamed_path(target):
    result = target.renamed_path()
    assert result == "/asdf/foo/bar - HEVC.asdf"


def test_in_place_file_path(target):
    result = target.in_place_file_path()
    assert result == "/asdf/foo/bar.asdf"


@patch("shutil.move")
@patch("os.remove")
def test_move_output_video(mock_remove, mock_move, target):
    target.temp_file = mock_temp_file("/foo/bar/baz")
    target._move_output_video()
    mock_move.assert_called_with("/foo/bar/baz", "/asdf/foo/bar - HEVC.asdf")
    mock_remove.assert_not_called()


@patch("shutil.move")
@patch("os.remove")
def test_move_output_video_dry_run(mock_remove, mock_move, target):
    target.dry_run = True
    target.temp_file = mock_temp_file("/foo/bar/baz")
    target._move_output_video()
    mock_move.assert_not_called()
    mock_remove.assert_not_called()


@patch("shutil.move")
@patch("os.remove")
def test_move_output_video_in_place(mock_remove, mock_move, target):
    target.temp_file = mock_temp_file("/foo/bar/baz")
    target.in_place = True
    target._move_output_video()
    mock_move.assert_any_call("/foo/bar/baz", "/asdf/foo/bar - HEVC.asdf")
    mock_remove.assert_called_with("/asdf/foo/bar.mkv")
    mock_move.assert_any_call("/asdf/foo/bar - HEVC.asdf", "/asdf/foo/bar.asdf")


@patch("shutil.move")
@patch("os.remove")
def test_move_output_video_in_place_dry_run(mock_remove, mock_move, target):
    target.dry_run = True
    target.temp_file = mock_temp_file("/foo/bar/baz")
    target.in_place = True
    target._move_output_video()
    mock_move.assert_not_called()
    mock_remove.assert_not_called()


def test_in_desired_format(target):
    target.video.codec = Codec("HEVC")
    target.video_settings.codec = Codec("HEVC")
    response = target.process()
    assert response == Status.IN_DESIRED_FORMAT


@patch.object(VideoProcessor, "already_processed", return_value=True)
def test_already_processed(mock_already_processed, target):
    response = target.process()
    assert response == Status.ALREADY_PROCESSED


@patch.object(VideoProcessor, "already_processed", return_value=False)
@patch.object(VideoProcessor, "_create_temp_file")
@patch("convert_videos.video_processor.FFmpegConverter")
@patch.object(VideoProcessor, "_move_output_video")
def test_converted(m1, m2, m3, m4, target):
    response = target.process()
    assert response == Status.CONVERTED


@patch.object(VideoProcessor, "already_processed", return_value=False)
@patch.object(VideoProcessor, "_create_temp_file")
@patch("convert_videos.video_processor.FFmpegConverter", side_effect=Exception)
@patch.object(VideoProcessor, "_move_output_video")
def test_failed(m1, m2, m3, m4, target):
    response = target.process()
    assert response == Status.FAILED


@patch.object(VideoProcessor, "_is_below_minimum_size", return_value=True)
def test_below_minimum_size(mock_is_below_minimum_size, target):
    response = target.process()
    assert response == Status.BELOW_MINIMUM_SIZE


def test_is_below_minimum_size_true(target):
    target.minimum_size_b = 200 * 1024 * 1024
    target.video.size_b = 100 * 1024 * 1024
    result = target._is_below_minimum_size()
    assert result is True


@patch("os.path.getsize", return_value=300 * 1024 * 1024)
def test_is_below_minimum_size_false(mock_getsize, target):
    target.minimum_size = 200
    result = target._is_below_minimum_size()
    assert result is False


@patch("os.path.getsize", return_value=300 * 1024 * 1024)
def test_is_below_minimum_size_zero(mock_getsize, target):
    target.minimum_size = 0
    result = target._is_below_minimum_size()
    assert result is False



# flake8: noqa: F401
from .processor import Processor
from .video_processor import VideoProcessor
from .settings import AudioSettings, VideoSettings
from .ffmpeg_converter import FFmpegConverter
from .util import check_hardware_acceleration_support



import logging
import os
from dataclasses import dataclass

import ffmpy

from .colour import colour
from .settings import AudioSettings, VideoSettings

log = logging.getLogger(__name__)


@dataclass
class FFmpegConverter:
    source_file_path: str
    destination_file_path: str
    extra_ffmpeg_input_args: str
    extra_ffmpeg_output_args: str
    dry_run: bool
    video_settings: VideoSettings
    audio_settings: AudioSettings

    def __post_init__(self):
        self._validate_destination()

    def process(self):
        input_settings = self._generate_ffmpeg_settings("input")
        output_settings = self._generate_ffmpeg_settings("output")

        ff = ffmpy.FFmpeg(
            inputs={self.source_file_path: input_settings},
            outputs={self.destination_file_path: output_settings},
        )
        if self.dry_run:
            log.info(
                colour("blue", f"DRY-RUN: Would start conversion. Command: '{ff.cmd}'")
            )
        else:
            log.info(colour("blue", f"Starting conversion. Command: '{ff.cmd}'"))
            ff.run()
            log.info(colour("green", "Successfully finished conversion!"))

    def _generate_ffmpeg_settings(self, mode):
        if mode == "input":
            output = " " + self.extra_ffmpeg_input_args
            if self.video_settings.encoder == "nvidia":
                output = " -hwaccel cuda"
            if self.video_settings.encoder == "intel":
                output = " -hwaccel qsv -hwaccel_output_format qsv"
            return output

        output = (
            ""
            + "-y"
            + " -threads 0"
            + str(self.video_settings)
            + str(self.audio_settings)
            + " "
            + self.extra_ffmpeg_output_args
        )
        return output

    def _validate_destination(self):
        if os.path.exists(self.destination_file_path):
            if os.path.isfile(self.destination_file_path):
                if os.access(self.destination_file_path, os.W_OK):
                    return True
                raise IOError(
                    f"Destination file {self.destination_file_path} exists but is not writable"
                )
            if os.path.isdir(self.destination_file_path):
                raise IOError(
                    f"Destination file {self.destination_file_path} is a directory, and not a file"
                )
        else:
            try:
                with open(self.destination_file_path, "w") as f:
                    f.writelines("test")
            finally:
                os.remove(self.destination_file_path)
            return True



import subprocess


def check_hardware_acceleration_support():
    support = {"intel_quicksync": False, "nvidia_nvenc": False}

    lsmod_output = subprocess.run(["lsmod"], capture_output=True, text=True)

    if "i915" in lsmod_output.stdout:
        support["intel_quicksync"] = True
    if "nvidia" in lsmod_output.stdout:
        support["nvidia_nvenc"] = True

    return support
